/**
 * Kernel - Brains of the operation
 *
 * @author	Jonathan Roewen
 */
#include "Kernel.h"

#include <fstream>
#include <iostream>
#include <strstream>

#include "Match.h"
#include "Utils.h"
#include "Memory.h"
#include "Parser.h"
#include "Element.h"
#include "Handler.h"
#include "Responder.h"
#include "Nodemapper.h"
#include "Nodemaster.h"
#include "Substituter.h"
#include "StringTokenizer.h"
#include "TemplateProcessor.h"

int totalCategories = 0;
long totalSize = 0;
long totalNodes = 0;

string toString(PElement element) {
	if (element == PElement()) {
		return "*";
	}
	string result;
	PElement child = element->getChild();
	while (child != PElement()) {
		if (child->getTagname() == "#text") {
			//	Then this child only contains text, no markup
			result += child->getText();
		} else {
			//	Since only using on <that> and <pattern>
			//	The tag can only be <bot_xxx/> or <bot name="xxx"/> or <name/>
			if (child->getTagname() == "name") {
				result += Kernel::respond("BOT NAME", "unknown");
			} else
			if (child->getTagname() == "bot") {
				//	Is a <bot name="xxx"/> predicate
				result += Kernel::respond("BOT " + toUpper(child->getAttribute("name")), "unknown");
			} else {
				//	It is old-skool <bot_xxx/>
				result += Kernel::respond("BOT " + toUpper(child->getTagname().substr(4)), "unknown");
			}
		}
		child = child->getNextSibling();
	}
	return trim(toUpper(result));//Substituter::normalize(result);
}

string templateString(PElement element) {
	if (element == NULL) {
		return "";
	}
	string result;
	PElement child = element->getChild();
	while (child != NULL) {
		if (child->getTagname() == "#text") {
			result += child->getText();
		} else {
			string tag = "<";
			if (child->hasNamespace()) {
				tag += child->getNamespace() + ":";
			}
			tag += child->getTagname();
			if (child->hasAttributes()) {
				map<string, string> attr = child->getAttributes();
				map<string, string>::const_iterator itr = attr.begin();
				while (itr != attr.end()) {
					tag += " " + (*itr).first + "=\"" + (*itr).second + "\"";
					itr++;
				}
			}
			if (child->hasChildren()) {
				tag += ">" + templateString(child);
				tag += "</";
				if (child->hasNamespace()) {
					tag += child->getNamespace() + ":";
				}
				tag += child->getTagname() + ">";
			} else {
				tag += "/>";
			}
			result += tag;
		}
		child = child->getNextSibling();
	}
	return result;
}

Nodemapper *Kernel::root = new Nodemaster();

Kernel::Kernel() {
	new Substituter();
	Handler::init();
	Kernel::load("std-startup.aiml");
}

void Kernel::load(const string &file) {
	Parser *p = new Parser();
	bool loadedOk = true;
	int total = totalCategories;
	ifstream fin;
	fin.open(file.c_str());
	if (fin.is_open()) {
		PElement root;
		root = p->parse(fin, file);
		if (root == PElement()) {
			loadedOk = false;
			cout << "Failed to load " << file << endl;
			return;
		}
		if (root->getTagname() == "aiml" || root->getTagname() == "alice") {
			load10(file, root);
		} else {
			loadedOk = false;
			cout << "Unexpected root element: " << root->getTagname() << endl;
		}
		fin.close();
		//	Need to cleanup .. deleting the parsed tree
	//	delete root;
	} else {
		loadedOk = false;
		cout << "File " << file << " could not be found" << endl;
	}
	if (loadedOk) {
		cout << "Learnt: " << file << " (" << (totalCategories - total) << " categories) (" << totalCategories << " total)" << endl;
	}
	delete p;
}

/**
 * At A & B we have a problem. Need to fix Element to do a
 * deep copy, so that we can safely delete the parsed tree.
 * Otherwise will crash and burn, or use far more memory
 * than is necessary.
 */
void Kernel::load10(const string &file, PElement &root) {
	velement array;
	unsigned int ix;
	
	root->getChildren("substitute", &array);
	for (ix = 0; ix < array.size(); ++ix) {
		string name = array[ix]->getAttribute("name");
		string find = array[ix]->getAttribute("find");
		string replace = array[ix]->getAttribute("replace");
		Substituter::addSubstitute(name, find, replace);
	}
	array.clear();
	
	root->getChildren("processor", &array);
	for (ix = 0; ix < array.size(); ++ix) {
		string tag = array[ix]->getAttribute("tag");
		string loc = array[ix]->getAttribute("class");
		if (Handler::hasProcessor(tag)) {
//	TEMPORARY WORKAROUND
//	Once have properly handled platform specific dynamic
//	linking, will make processors dynamically linked in
//	at startup, rather than statically provided in Handler
			TemplateProcessor::addProcessor((*Handler::processors.find(tag)).second);
		//	cout << "Adding Processor: " << loc << endl;
		} else {
		//	cout << "No " << tag << " processor" << endl;
		}
	}
	array.clear();
	
//	Now to actually load the AIML itself...
	root->getChildren("topic", &array);
	//	All the <topic> nodes
	for (ix = 0; ix < array.size(); ++ix) {
		velement categories;
		array[ix]->getChildren("category", &categories);
		//	All of the <category> nodes per <topic>
		string topic = array[ix]->getAttribute("name");
		for (unsigned int iy = 0; iy < categories.size(); ++iy) {
			//	Load each category
			string that = toString(categories[iy]->getChild("that"));
			string pattern = toString(categories[iy]->getChild("pattern"));
			Nodemapper *node = add(pattern, that, topic);
			if (node != NULL) {
				node->addTemplate(templateString(categories[iy]->getChild("template")));
				node->addAuthor("j-alice");
				node->addFilename(file);
				++totalCategories;
			}
		}
	}
	array.clear();
	
	root->getChildren("category", &array);
	//	All the <category> nodes without a <topic>
	for (ix = 0; ix < array.size(); ++ix) {
		//	Load each category
		string that = toString(array[ix]->getChild("that"));
		string pattern = toString(array[ix]->getChild("pattern"));
		Nodemapper *node = add(pattern, that, "*");
		if (node != NULL) {
			node->addTemplate(templateString(array[ix]->getChild("template")));
			node->addAuthor("j-alice");
			node->addFilename(file);
			++totalCategories;
		}
	}
//	And I think that concludes loading the AIML :-)
}

Nodemapper *Kernel::add(const string &pattern, const string &that, const string &topic) {
	string path = pattern + " <that> " + that + " <topic> " + topic;
	Nodemapper *n = add(path, root);
	return n;
}

Nodemapper *Kernel::add(const string &sentence, Nodemapper *parent) {
	StringTokenizer st(sentence, " ");
	int count = st.countTokens();
	Nodemapper *n = add(st, parent, count, sentence);
	return n;
}

Nodemapper *Kernel::add(StringTokenizer &st, Nodemapper *parent, int count, const string &sentence) {
	if (count == 0) {
		return parent;
	} else {
		string word = st.nextToken();
		Nodemapper *n = parent->getChild(word);
		if (n == NULL) {
//	new Nodemapper()? It's an ABC...
			n = new Nodemaster();
			parent->addChild(word, n);
		}
		return add(st, n, count - 1, sentence);
	}
}

Match *Kernel::match(const string &input, const string &that, const string &topic) {
	Match *m = NULL;
	string path = input + " <that> " + that + " <topic> " + topic;
//	cout << path << endl;
	m = match(root, root, path, "", "");
	if (m != NULL) {
		m->setPath(m->getInputPattern() + " : " + m->getThatPattern() + " : " + m->getTopicPattern());
		return m;
	} else {
		throw(int(0));
	}
}

/**
 * Needs optimizing .. perhaps changes to the Nodemaster/Nodemapper
 * will have to be included.
 * For example: input of say 5 words, and 24 wildcard pattern is
 * extremely slow, and will dramatically increase in time to
 * complete as graphmaster size increases.
 *
 * If (max_tokens(pattern) > max_tokens(remaining input))
 * then stop searching this branch.
 *
 * Still, this is not fail proof, but it will lead to faster
 * running times in the average case :-)
 */
Match *Kernel::match(Nodemapper *node, Nodemapper *parent, const string &input, const string &star, const string &path) {
//cout << input << " : " << star << " : " << path << endl;
	StringTokenizer st(input, " ");
	int count = st.countTokens();
	Match *m = NULL;
	if (count == 0) {
		if (!node->getTemplate().empty()) {
			m = new Match();
			m->setNode(node);
			//	Is this correct?
			m->addTopicStar(trim(star));
			m->setTopicPattern("*");
			return m;
		} else {
			return NULL;
		}
	}
	string word = st.nextToken();
	string tail = "";
	if (st.hasMoreTokens()) {
		tail = input.substr(word.length() + 1, input.length());
	}
	if (node->getChild("_") != NULL) {
		m = match(node->getChild("_"), node, tail, word, path + " _");
		if (m != NULL) {
			if (!trim(star).empty()) {
				m->addInputStar(trim(star));
			}
			return m;
		}
//		cout << "No _ (" << path << ")" << endl;
	}
	if (node->getChild(word) != NULL) {
		if (word[0] == '<') {
			m = match(node->getChild(word), node, tail, star, "");
		} else {
			m = match(node->getChild(word), node, tail, star, path + " " + word);
		}
		if (m != NULL) {
			if (word == "<that>") {
				m->addInputStar(trim(star));
				m->setInputPattern(trim(path));
			} else
			if (word == "<topic>") {
				m->addThatStar(trim(star));
				m->setThatPattern(trim(path));
			}
			return m;
		}
//		cout << "No " << word << " (" << path << ")" << endl;
	}
	if (node->getChild("*") != NULL) {
		m = match(node->getChild("*"), node, tail, word, path + " *");
		if (m != NULL) {
			if (!trim(star).empty()) {
				m->addInputStar(trim(star));
			}
			return m;
		}
//		cout << "No * (" << path << ")" << endl;
	}
	if (node == parent->getChild("*") || node == parent->getChild("_")) {
//		cout << "Adding word to <star/>" << endl;
		return match(node, parent, tail, star + " " + word, path);
	}
	return NULL;
}

string Kernel::process(Match *m, PElement e, Responder *r, const string &id) {
	return TemplateProcessor::processTemplate(m, e, r, id);
}

string Kernel::respond(const string &input, const string &id) {
	return respond(input, id, 1, false);
}

string Kernel::respond(const string &input, const string &id, bool srai) {
	return respond(input, id, 1, srai);
}

string Kernel::respond(const string &input, const string &id, Responder *r) {
	return respond(input, id, r, 1, false);
}

string Kernel::respond(const string &input, const string &id, int depth, bool srai) {
	/**
	 * Instead of NULL, should be: new TextResponder() or something.
	 * NULL isn't going to work later on
	 */
	return respond(input, id, NULL, depth, srai);
}

string Kernel::respond(const string &input, const string &id, Responder *r, int depth, bool srai) {
//	cout << "KERNEL: " << input << ":" << id << endl;
	if (depth > 128) {
		Memory::push("that", "", id);
		return "I am confused .. sorry :-(";
	}
	string currentResponse, buffer = "";
	Match *m = NULL;
	if (!srai) {
		Memory::push("input", input, id);
	}
	StringTokenizer st(input, ".?!");
	while (st.hasMoreTokens()) {
		string sentence = st.nextToken();
		sentence = Substituter::normalize(sentence);
		string that = Memory::getValue("that", id);
		if (!srai) {
			StringTokenizer stThat(that, ".?!");
			while (stThat.hasMoreTokens()) {
				that = stThat.nextToken();
				that = trim(that);
				Memory::push("that", that, id);
			}
		}
		that = Substituter::normalize(that);
		string topic = Memory::getValue("topic", id);
		topic = Substituter::normalize(topic);
		if (that.empty()) {
			that = "*";
		}
		if (topic.empty()) {
			topic = "*";
		}
		try {
//			cout << "MATCH: " << sentence << " : " << that << " : " << topic << endl;
			m = match(sentence, that, topic);
			//	DAMMIT! DO I NEED HERE AS WELL?????
			string tmpl = "<template>" + m->getTemplate() + "</template>";
			Parser *p = new Parser();
			strstream ss;
			ss << tmpl << endl;
			currentResponse = Kernel::process(m, p->parse(ss, "xxx"), r, id);
			Memory::setValue("beforethat", id, that);
			Memory::setValue("that", id, currentResponse);
			delete p;
		} catch (int &ex) {
		//	buffer += "[Error - No Match] ";
		}
		if (m != NULL) {
		//	cout << "PATH: " << m->getPath() << endl;
			delete m;
			if (srai) {
				return currentResponse;
			} else {
				buffer += currentResponse + " ";
			}
		}
	}
	string result = Substituter::output(buffer);
//	cout << totalSize << " : " << totalNodes << endl;
//	cout << "Tree Size: " << (totalNodes * (sizeof(Nodemaster) + sizeof(map<string, Nodemaster *>) + sizeof(int) + sizeof(string))) << endl;
	return trim(result, " \t\r\n");
}